@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2017 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -50,8 +50,28 @@ var (
 	pollTime    = 100 * time.Millisecond
 )
 
+type waitMixin struct {
+	NoWait bool `long:"no-wait" hidden:"true"`
+}
+
+// TODO: use waitMixin outside of cmd_snap_op.go
+
+var waitDescs = mixinDescs{
+	"no-wait": i18n.G("Do not wait for the operation to finish but just print the change id."),
+}
+
+var noWait = errors.New("no wait for op")
+
+func (wmx *waitMixin) wait(cli *client.Client, id string) (*client.Change, error) {
+	if wmx.NoWait {
+		fmt.Fprintf(Stdout, "%s\n", id)
+		return nil, noWait
+	}
+	return wait(cli, id)
+}
+
 func wait(cli *client.Client, id string) (*client.Change, error) {
-	pb := progress.NewTextProgress()
+	pb := progress.MakeProgressBar()
 	defer func() {
 		pb.Finished()
 	}()
@@ -102,7 +122,7 @@ func wait(cli *client.Client, id string) (*client.Change, error) {
 			case t.ID == lastID:
 				pb.Set(float64(t.Progress.Done))
 			default:
-				pb.Start(t.Progress.Label, float64(t.Progress.Total))
+				pb.Start(t.Summary, float64(t.Progress.Total))
 				lastID = t.ID
 			}
 			break
@@ -174,6 +194,8 @@ and the snap can easily be enabled again.
 `)
 
 type cmdRemove struct {
+	waitMixin
+
 	Revision   string `long:"revision"`
 	Positional struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>" required:"1"`
@@ -181,39 +203,43 @@ type cmdRemove struct {
 }
 
 func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
-	name := x.Positional.Snaps[0]
+	name := string(x.Positional.Snaps[0])
 
 	cli := Client()
-	changeID, err := cli.Remove(string(name), opts)
-	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindSnapNotInstalled {
-		fmt.Fprintf(Stderr, e.Message+"\n")
+	changeID, err := cli.Remove(name, opts)
+	if err != nil {
+		msg, err := errorToCmdMessage(name, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
+	}
+
+	_, err = x.wait(cli, changeID)
+	if err == noWait {
 		return nil
 	}
 	if err != nil {
 		return err
 	}
 
-	if _, err := wait(cli, changeID); err != nil {
-		return err
-	}
-
 	fmt.Fprintf(Stdout, i18n.G("%s removed\n"), name)
 	return nil
 }
 
 func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
-	names := make([]string, len(x.Positional.Snaps))
-	for i, s := range x.Positional.Snaps {
-		names[i] = string(s)
-	}
-
+	names := installedSnapNames(x.Positional.Snaps)
 	cli := Client()
 	changeID, err := cli.RemoveMany(names, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := wait(cli, changeID)
+	chg, err := x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -302,6 +328,12 @@ func (mx *channelMixin) setChannelFromCommandline() error {
 		mx.Channel = ch.chName
 	}
 
+	if !strings.Contains(mx.Channel, "/") && mx.Channel != "" && mx.Channel != "edge" && mx.Channel != "beta" && mx.Channel != "candidate" && mx.Channel != "stable" {
+		// shortcut to jump to a different track, e.g.
+		// snap install foo --channel=3.4 # implies 3.4/stable
+		mx.Channel += "/stable"
+	}
+
 	return nil
 }
 
@@ -321,20 +353,32 @@ func showDone(names []string, op string) error {
 		switch op {
 		case "install":
 			if snap.Developer != "" {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' installed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' installed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
 			} else {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 installed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s installed\n"), snap.Name, channelStr, snap.Version)
 			}
 		case "refresh":
 			if snap.Developer != "" {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' refreshed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' refreshed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
 			} else {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 refreshed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s refreshed\n"), snap.Name, channelStr, snap.Version)
 			}
+		case "revert":
+			// TRANSLATORS: first %s is a snap name, second %s is a revision
+			fmt.Fprintf(Stdout, i18n.G("%s reverted to %s\n"), snap.Name, snap.Version)
 		default:
-			fmt.Fprintf(Stdout, "internal error, unknown op %q", op)
+			fmt.Fprintf(Stdout, "internal error: unknown op %q", op)
+		}
+		if snap.TrackingChannel != snap.Channel {
+			// TRANSLATORS: first %s is a snap name, following %s is a channel name
+			fmt.Fprintf(Stdout, i18n.G("Snap %s is no longer tracking %s.\n"), snap.Name, snap.TrackingChannel)
 		}
 	}
+
 	return nil
 }
 
@@ -374,6 +418,8 @@ func (mx modeMixin) setModes(opts *client.SnapOptions) {
 }
 
 type cmdInstall struct {
+	waitMixin
+
 	channelMixin
 	modeMixin
 	Revision string `long:"revision"`
@@ -383,6 +429,8 @@ type cmdInstall struct {
 	// because we released 2.14.2 with --force-dangerous
 	ForceDangerous bool `long:"force-dangerous" hidden:"yes"`
 
+	Unaliased bool `long:"unaliased"`
+
 	Positional struct {
 		Snaps []remoteSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
@@ -414,17 +462,21 @@ func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
 	} else {
 		changeID, err = cli.Install(name, opts)
 	}
-	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindSnapAlreadyInstalled {
-		fmt.Fprintf(Stderr, i18n.G("snap %q is already installed, see \"snap refresh --help\"\n"), name)
-		return nil
-	}
 	if err != nil {
-		return err
+		msg, err := errorToCmdMessage(name, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
 	}
 
 	setupAbortHandler(changeID)
 
-	chg, err := wait(cli, changeID)
+	chg, err := x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -453,12 +505,24 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 	cli := Client()
 	changeID, err := cli.InstallMany(names, opts)
 	if err != nil {
-		return err
+		var snapName string
+		if err, ok := err.(*client.Error); ok {
+			snapName, _ = err.Value.(string)
+		}
+		msg, err := errorToCmdMessage(snapName, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
 	}
 
 	setupAbortHandler(changeID)
 
-	chg, err := wait(cli, changeID)
+	chg, err := x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -503,14 +567,11 @@ func (x *cmdInstall) Execute([]string) error {
 		Channel:   x.Channel,
 		Revision:  x.Revision,
 		Dangerous: dangerous,
+		Unaliased: x.Unaliased,
 	}
 	x.setModes(opts)
 
-	names := make([]string, len(x.Positional.Snaps))
-	for i, name := range x.Positional.Snaps {
-		names[i] = string(name)
-	}
-
+	names := remoteSnapNames(x.Positional.Snaps)
 	if len(names) == 1 {
 		return x.installOne(names[0], opts)
 	}
@@ -523,25 +584,32 @@ func (x *cmdInstall) Execute([]string) error {
 }
 
 type cmdRefresh struct {
+	waitMixin
+
 	channelMixin
 	modeMixin
 
+	Amend            bool   `long:"amend"`
 	Revision         string `long:"revision"`
 	List             bool   `long:"list"`
+	Time             bool   `long:"time"`
 	IgnoreValidation bool   `long:"ignore-validation"`
 	Positional       struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes"`
 }
 
-func refreshMany(snaps []string, opts *client.SnapOptions) error {
+func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error {
 	cli := Client()
 	changeID, err := cli.RefreshMany(snaps, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := wait(cli, changeID)
+	chg, err := x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -560,25 +628,60 @@ func refreshMany(snaps []string, opts *client.SnapOptions) error {
 	return nil
 }
 
-func refreshOne(name string, opts *client.SnapOptions) error {
+func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
 	cli := Client()
 	changeID, err := cli.Refresh(name, opts)
-	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindNoUpdateAvailable {
-		fmt.Fprintf(Stderr, e.Message+"\n")
+	if err != nil {
+		msg, err := errorToCmdMessage(name, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
+	}
+
+	_, err = x.wait(cli, changeID)
+	if err == noWait {
 		return nil
 	}
 	if err != nil {
 		return err
 	}
 
-	if _, err := wait(cli, changeID); err != nil {
-		return err
-	}
-
 	return showDone([]string{name}, "refresh")
 }
 
-func listRefresh() error {
+func (x *cmdRefresh) showRefreshTimes() error {
+	cli := Client()
+	sysinfo, err := cli.SysInfo()
+	if err != nil {
+		return err
+	}
+
+	if sysinfo.Refresh.Timer != "" {
+		fmt.Fprintf(Stdout, "timer: %s\n", sysinfo.Refresh.Timer)
+	} else if sysinfo.Refresh.Schedule != "" {
+		fmt.Fprintf(Stdout, "schedule: %s\n", sysinfo.Refresh.Schedule)
+	} else {
+		return errors.New("internal error: both refresh.timer and refresh.schedule are empty")
+	}
+	if sysinfo.Refresh.Last != "" {
+		fmt.Fprintf(Stdout, "last: %s\n", sysinfo.Refresh.Last)
+	} else {
+		fmt.Fprintf(Stdout, "last: n/a\n")
+	}
+	if sysinfo.Refresh.Hold != "" {
+		fmt.Fprintf(Stdout, "hold: %s\n", sysinfo.Refresh.Hold)
+	}
+	if sysinfo.Refresh.Next != "" {
+		fmt.Fprintf(Stdout, "next: %s\n", sysinfo.Refresh.Next)
+	} else {
+		fmt.Fprintf(Stdout, "next: n/a\n")
+	}
+	return nil
+}
+
+func (x *cmdRefresh) listRefresh() error {
 	cli := Client()
 	snaps, _, err := cli.Find(&client.FindOptions{
 		Refresh: true,
@@ -612,25 +715,36 @@ func (x *cmdRefresh) Execute([]string) error {
 		return err
 	}
 
+	if x.Time {
+		if x.asksForMode() || x.asksForChannel() {
+			return errors.New(i18n.G("--time does not take mode nor channel flags"))
+		}
+		return x.showRefreshTimes()
+	}
+
 	if x.List {
 		if x.asksForMode() || x.asksForChannel() {
 			return errors.New(i18n.G("--list does not take mode nor channel flags"))
 		}
 
-		return listRefresh()
+		return x.listRefresh()
 	}
-	names := make([]string, len(x.Positional.Snaps))
-	for i, name := range x.Positional.Snaps {
-		names[i] = string(name)
+
+	if len(x.Positional.Snaps) == 0 && os.Getenv("SNAP_REFRESH_FROM_TIMER") == "1" {
+		fmt.Fprintf(Stdout, "Ignoring `snap refresh` from the systemd timer")
+		return nil
 	}
-	if len(x.Positional.Snaps) == 1 {
+
+	names := installedSnapNames(x.Positional.Snaps)
+	if len(names) == 1 {
 		opts := &client.SnapOptions{
+			Amend:            x.Amend,
 			Channel:          x.Channel,
 			IgnoreValidation: x.IgnoreValidation,
 			Revision:         x.Revision,
 		}
 		x.setModes(opts)
-		return refreshOne(names[0], opts)
+		return x.refreshOne(names[0], opts)
 	}
 
 	if x.asksForMode() || x.asksForChannel() {
@@ -641,16 +755,32 @@ func (x *cmdRefresh) Execute([]string) error {
 		return errors.New(i18n.G("a single snap name must be specified when ignoring validation"))
 	}
 
-	return refreshMany(names, nil)
+	return x.refreshMany(names, nil)
 }
 
 type cmdTry struct {
+	waitMixin
+
 	modeMixin
 	Positional struct {
 		SnapDir string `positional-arg-name:"<snap-dir>"`
 	} `positional-args:"yes"`
 }
 
+func hasSnapcraftYaml() bool {
+	for _, loc := range []string{
+		"snap/snapcraft.yaml",
+		"snapcraft.yaml",
+		".snapcraft.yaml",
+	} {
+		if osutil.FileExists(loc) {
+			return true
+		}
+	}
+
+	return false
+}
+
 func (x *cmdTry) Execute([]string) error {
 	if err := x.validateMode(); err != nil {
 		return err
@@ -661,7 +791,7 @@ func (x *cmdTry) Execute([]string) error {
 	x.setModes(opts)
 
 	if name == "" {
-		if osutil.FileExists("snapcraft.yaml") && osutil.IsDirectory("prime") {
+		if hasSnapcraftYaml() && osutil.IsDirectory("prime") {
 			name = "prime"
 		} else {
 			if osutil.FileExists("meta/snap.yaml") {
@@ -680,11 +810,19 @@ func (x *cmdTry) Execute([]string) error {
 	}
 
 	changeID, err := cli.Try(path, opts)
+	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindNotSnap {
+		return fmt.Errorf(i18n.G(`%q does not contain an unpacked snap.
+
+Try "snapcraft prime" in your project directory, then "snap try" again.`), path)
+	}
 	if err != nil {
 		return err
 	}
 
-	chg, err := wait(cli, changeID)
+	chg, err := x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -713,6 +851,8 @@ func (x *cmdTry) Execute([]string) error {
 }
 
 type cmdEnable struct {
+	waitMixin
+
 	Positional struct {
 		Snap installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
@@ -727,7 +867,10 @@ func (x *cmdEnable) Execute([]string) error {
 		return err
 	}
 
-	_, err = wait(cli, changeID)
+	_, err = x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -737,6 +880,8 @@ func (x *cmdEnable) Execute([]string) error {
 }
 
 type cmdDisable struct {
+	waitMixin
+
 	Positional struct {
 		Snap installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
@@ -751,7 +896,10 @@ func (x *cmdDisable) Execute([]string) error {
 		return err
 	}
 
-	_, err = wait(cli, changeID)
+	_, err = x.wait(cli, changeID)
+	if err == noWait {
+		return nil
+	}
 	if err != nil {
 		return err
 	}
@@ -761,11 +909,13 @@ func (x *cmdDisable) Execute([]string) error {
 }
 
 type cmdRevert struct {
+	waitMixin
+
 	modeMixin
 	Revision   string `long:"revision"`
 	Positional struct {
 		Snap installedSnapName `positional-arg-name:"<snap>"`
-	} `positional-args:"yes"`
+	} `positional-args:"yes" required:"yes"`
 }
 
 var shortRevertHelp = i18n.G("Reverts the given snap to the previous state")
@@ -796,43 +946,82 @@ func (x *cmdRevert) Execute(args []string) error {
 		return err
 	}
 
-	if _, err := wait(cli, changeID); err != nil {
-		return err
+	_, err = x.wait(cli, changeID)
+	if err == noWait {
+		return nil
 	}
-
-	// show output as speced
-	snaps, err := cli.List([]string{name}, nil)
 	if err != nil {
 		return err
 	}
-	if len(snaps) != 1 {
-		// TRANSLATORS: %q gets the snap name, %v the list of things found when trying to list it
-		return fmt.Errorf(i18n.G("cannot get data for %q: %v"), name, snaps)
+
+	return showDone([]string{name}, "revert")
+}
+
+var shortSwitchHelp = i18n.G("Switches snap to a different channel")
+var longSwitchHelp = i18n.G(`
+The switch command switches the given snap to a different channel without
+doing a refresh.
+`)
+
+type cmdSwitch struct {
+	channelMixin
+
+	Positional struct {
+		Snap installedSnapName `positional-arg-name:"<snap>" required:"1"`
+	} `positional-args:"yes" required:"yes"`
+}
+
+func (x cmdSwitch) Execute(args []string) error {
+	if err := x.setChannelFromCommandline(); err != nil {
+		return err
 	}
-	snap := snaps[0]
-	fmt.Fprintf(Stdout, i18n.G("%s reverted to %s\n"), name, snap.Version)
+	if x.Channel == "" {
+		return fmt.Errorf("missing --channel=<channel-name> parameter")
+	}
+
+	cli := Client()
+	name := string(x.Positional.Snap)
+	channel := string(x.Channel)
+	opts := &client.SnapOptions{
+		Channel: channel,
+	}
+	changeID, err := cli.Switch(name, opts)
+	if err != nil {
+		return err
+	}
+
+	if _, err = wait(cli, changeID); err != nil {
+		return err
+	}
+
+	fmt.Fprintf(Stdout, i18n.G("%q switched to the %q channel\n"), name, channel)
 	return nil
 }
 
 func init() {
 	addCommand("remove", shortRemoveHelp, longRemoveHelp, func() flags.Commander { return &cmdRemove{} },
-		map[string]string{"revision": i18n.G("Remove only the given revision")}, nil)
+		waitDescs.also(map[string]string{"revision": i18n.G("Remove only the given revision")}), nil)
 	addCommand("install", shortInstallHelp, longInstallHelp, func() flags.Commander { return &cmdInstall{} },
-		channelDescs.also(modeDescs).also(map[string]string{
+		waitDescs.also(channelDescs).also(modeDescs).also(map[string]string{
 			"revision":        i18n.G("Install the given revision of a snap, to which you must have developer access"),
 			"dangerous":       i18n.G("Install the given snap file even if there are no pre-acknowledged signatures for it, meaning it was not verified and could be dangerous (--devmode implies this)"),
 			"force-dangerous": i18n.G("Alias for --dangerous (DEPRECATED)"),
+			"unaliased":       i18n.G("Install the given snap without enabling its automatic aliases"),
 		}), nil)
 	addCommand("refresh", shortRefreshHelp, longRefreshHelp, func() flags.Commander { return &cmdRefresh{} },
-		channelDescs.also(modeDescs).also(map[string]string{
+		waitDescs.also(channelDescs).also(modeDescs).also(map[string]string{
+			"amend":             i18n.G("Allow refresh attempt on snap unknown to the store"),
 			"revision":          i18n.G("Refresh to the given revision"),
-			"list":              i18n.G("Show available snaps for refresh"),
+			"list":              i18n.G("Show available snaps for refresh but do not perform a refresh"),
+			"time":              i18n.G("Show auto refresh information but do not perform a refresh"),
 			"ignore-validation": i18n.G("Ignore validation by other snaps blocking the refresh"),
 		}), nil)
-	addCommand("try", shortTryHelp, longTryHelp, func() flags.Commander { return &cmdTry{} }, modeDescs, nil)
-	addCommand("enable", shortEnableHelp, longEnableHelp, func() flags.Commander { return &cmdEnable{} }, nil, nil)
-	addCommand("disable", shortDisableHelp, longDisableHelp, func() flags.Commander { return &cmdDisable{} }, nil, nil)
-	addCommand("revert", shortRevertHelp, longRevertHelp, func() flags.Commander { return &cmdRevert{} }, modeDescs.also(map[string]string{
+	addCommand("try", shortTryHelp, longTryHelp, func() flags.Commander { return &cmdTry{} }, waitDescs.also(modeDescs), nil)
+	addCommand("enable", shortEnableHelp, longEnableHelp, func() flags.Commander { return &cmdEnable{} }, waitDescs, nil)
+	addCommand("disable", shortDisableHelp, longDisableHelp, func() flags.Commander { return &cmdDisable{} }, waitDescs, nil)
+	addCommand("revert", shortRevertHelp, longRevertHelp, func() flags.Commander { return &cmdRevert{} }, waitDescs.also(modeDescs).also(map[string]string{
 		"revision": "Revert to the given revision",
 	}), nil)
+	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, nil, nil)
+
 }
