@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2017 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -22,10 +22,13 @@ package snapstate_test
 import (
 	"errors"
 	"fmt"
+	"io"
+	"path/filepath"
+	"sort"
+	"strings"
 
 	"golang.org/x/net/context"
 
-	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/snapstate/backend"
@@ -33,6 +36,7 @@ import (
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
+	"github.com/snapcore/snapd/store/storetest"
 )
 
 type fakeOp struct {
@@ -49,6 +53,8 @@ type fakeOp struct {
 
 	aliases   []*backend.Alias
 	rmAliases []*backend.Alias
+
+	userID int
 }
 
 type fakeOps []fakeOp
@@ -88,7 +94,10 @@ type fakeDownload struct {
 }
 
 type fakeStore struct {
+	storetest.Store
+
 	downloads           []fakeDownload
+	refreshRevnos       map[string]snap.Revision
 	fakeBackend         *fakeSnappyBackend
 	fakeCurrentProgress int
 	fakeTotalProgress   int
@@ -113,12 +122,6 @@ func (f *fakeStore) SnapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.I
 	}
 
 	confinement := snap.StrictConfinement
-	switch spec.Channel {
-	case "channel-for-devmode":
-		confinement = snap.DevModeConfinement
-	case "channel-for-classic":
-		confinement = snap.ClassicConfinement
-	}
 
 	typ := snap.TypeApp
 	if spec.Name == "some-core" {
@@ -130,7 +133,7 @@ func (f *fakeStore) SnapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.I
 		SideInfo: snap.SideInfo{
 			RealName: spec.Name,
 			Channel:  spec.Channel,
-			SnapID:   "snapIDsnapidsnapidsnapidsnapidsn",
+			SnapID:   spec.Name + "-id",
 			Revision: spec.Revision,
 		},
 		Version: spec.Name,
@@ -140,89 +143,147 @@ func (f *fakeStore) SnapInfo(spec store.SnapSpec, user *auth.UserState) (*snap.I
 		Confinement: confinement,
 		Type:        typ,
 	}
-	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-snap", name: spec.Name, revno: spec.Revision})
+	switch spec.Channel {
+	case "channel-for-devmode":
+		info.Confinement = snap.DevModeConfinement
+	case "channel-for-classic":
+		info.Confinement = snap.ClassicConfinement
+	case "channel-for-paid":
+		info.Prices = map[string]float64{"USD": 0.77}
+		info.SideInfo.Paid = true
+	case "channel-for-private":
+		info.SideInfo.Private = true
+	case "channel-for-layout":
+		info.Layout = map[string]*snap.Layout{
+			"/usr": {
+				Snap:    info,
+				Path:    "/usr",
+				Symlink: "$SNAP/usr",
+			},
+		}
+	}
+
+	userID := 0
+	if user != nil {
+		userID = user.ID
+	}
+	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-snap", name: spec.Name, revno: spec.Revision, userID: userID})
 
 	return info, nil
 }
 
-func (f *fakeStore) Find(search *store.Search, user *auth.UserState) ([]*snap.Info, error) {
-	panic("Find called")
+func (f *fakeStore) LookupRefresh(cand *store.RefreshCandidate, user *auth.UserState) (*snap.Info, error) {
+	f.pokeStateLock()
+
+	if cand == nil {
+		panic("LookupRefresh called with no candidate")
+	}
+
+	var name string
+
+	switch cand.SnapID {
+	case "":
+		return nil, store.ErrLocalSnap
+	case "other-snap-id":
+		return nil, store.ErrNoUpdateAvailable
+	case "fakestore-please-error-on-refresh":
+		return nil, fmt.Errorf("failing as requested")
+	case "services-snap-id":
+		name = "services-snap"
+	case "some-snap-id":
+		name = "some-snap"
+	case "core-snap-id":
+		name = "core"
+	case "snap-with-snapd-control-id":
+		name = "snap-with-snapd-control"
+	default:
+		panic(fmt.Sprintf("ListRefresh: unknown snap-id: %s", cand.SnapID))
+	}
+
+	revno := snap.R(11)
+	if r := f.refreshRevnos[cand.SnapID]; !r.Unset() {
+		revno = r
+	}
+	confinement := snap.StrictConfinement
+	switch cand.Channel {
+	case "channel-for-7":
+		revno = snap.R(7)
+	case "channel-for-classic":
+		confinement = snap.ClassicConfinement
+	case "channel-for-devmode":
+		confinement = snap.DevModeConfinement
+	}
+
+	info := &snap.Info{
+		SideInfo: snap.SideInfo{
+			RealName: name,
+			Channel:  cand.Channel,
+			SnapID:   cand.SnapID,
+			Revision: revno,
+		},
+		Version: name,
+		DownloadInfo: snap.DownloadInfo{
+			DownloadURL: "https://some-server.com/some/path.snap",
+		},
+		Confinement:   confinement,
+		Architectures: []string{"all"},
+	}
+	switch cand.Channel {
+	case "channel-for-layout":
+		info.Layout = map[string]*snap.Layout{
+			"/usr": {
+				Snap:    info,
+				Path:    "/usr",
+				Symlink: "$SNAP/usr",
+			},
+		}
+	}
+
+	var hit snap.Revision
+	if cand.Revision != revno {
+		hit = revno
+	}
+	for _, blocked := range cand.Block {
+		if blocked == revno {
+			hit = snap.Revision{}
+			break
+		}
+	}
+
+	userID := 0
+	if user != nil {
+		userID = user.ID
+	}
+	// TODO: move this back to ListRefresh
+	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-list-refresh", cand: *cand, revno: hit, userID: userID})
+
+	if !hit.Unset() {
+		return info, nil
+	}
+
+	return nil, store.ErrNoUpdateAvailable
 }
 
-func (f *fakeStore) ListRefresh(cands []*store.RefreshCandidate, _ *auth.UserState) ([]*snap.Info, error) {
+func (f *fakeStore) ListRefresh(ctx context.Context, cands []*store.RefreshCandidate, user *auth.UserState, flags *store.RefreshOptions) ([]*snap.Info, error) {
+	if ctx == nil {
+		panic("context required")
+	}
 	f.pokeStateLock()
 
 	if len(cands) == 0 {
 		return nil, nil
 	}
-	if len(cands) > 2 {
-		panic("ListRefresh unexpectedly called with more than two candidates")
+	if len(cands) > 3 {
+		panic("fake ListRefresh unexpectedly called with more than 3 candidates")
 	}
 
 	var res []*snap.Info
 	for _, cand := range cands {
-		snapID := cand.SnapID
-
-		if snapID == "" || snapID == "other-snap-id" {
+		info, err := f.LookupRefresh(cand, user)
+		if err == store.ErrLocalSnap || err == store.ErrNoUpdateAvailable {
 			continue
 		}
-
-		if snapID == "fakestore-please-error-on-refresh" {
-			return nil, fmt.Errorf("failing as requested")
-		}
-
-		var name string
-		switch snapID {
-		case "some-snap-id":
-			name = "some-snap"
-		case "core-snap-id":
-			name = "core"
-		default:
-			panic(fmt.Sprintf("ListRefresh: unknown snap-id: %s", snapID))
-		}
-
-		revno := snap.R(11)
-		confinement := snap.StrictConfinement
-		switch cand.Channel {
-		case "channel-for-7":
-			revno = snap.R(7)
-		case "channel-for-classic":
-			confinement = snap.ClassicConfinement
-		case "channel-for-devmode":
-			confinement = snap.DevModeConfinement
-		}
-
-		info := &snap.Info{
-			SideInfo: snap.SideInfo{
-				RealName: name,
-				Channel:  cand.Channel,
-				SnapID:   cand.SnapID,
-				Revision: revno,
-			},
-			Version: name,
-			DownloadInfo: snap.DownloadInfo{
-				DownloadURL: "https://some-server.com/some/path.snap",
-			},
-			Confinement:   confinement,
-			Architectures: []string{"all"},
-		}
-
-		var hit snap.Revision
-		if cand.Revision != revno {
-			hit = revno
-		}
-		for _, blocked := range cand.Block {
-			if blocked == revno {
-				hit = snap.Revision{}
-				break
-			}
-		}
-
-		f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{op: "storesvc-list-refresh", cand: *cand, revno: hit})
-
-		if !hit.Unset() {
-			res = append(res, info)
-		}
+		res = append(res, info)
 	}
 
 	return res, nil
@@ -253,20 +314,28 @@ func (f *fakeStore) Download(ctx context.Context, name, targetFn string, snapInf
 	return nil
 }
 
-func (f *fakeStore) Buy(options *store.BuyOptions, user *auth.UserState) (*store.BuyResult, error) {
-	panic("Never expected fakeStore.Buy to be called")
+func (f *fakeStore) WriteCatalogs(ctx context.Context, _ io.Writer, _ store.SnapAdder) error {
+	if ctx == nil {
+		panic("context required")
+	}
+	f.pokeStateLock()
+	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{
+		op: "x-commands",
+	})
+
+	return nil
 }
 
-func (f *fakeStore) ReadyToBuy(user *auth.UserState) error {
-	panic("Never expected fakeStore.ReadyToBuy to be called")
-}
+func (f *fakeStore) Sections(ctx context.Context, _ *auth.UserState) ([]string, error) {
+	if ctx == nil {
+		panic("context required")
+	}
+	f.pokeStateLock()
+	f.fakeBackend.ops = append(f.fakeBackend.ops, fakeOp{
+		op: "x-sections",
+	})
 
-func (f *fakeStore) Assertion(*asserts.AssertionType, []string, *auth.UserState) (asserts.Assertion, error) {
-	panic("Never expected fakeStore.Assertion to be called")
-}
-
-func (f *fakeStore) Sections(user *auth.UserState) ([]string, error) {
-	panic("Sections called")
+	return nil, nil
 }
 
 type fakeSnappyBackend struct {
@@ -274,6 +343,7 @@ type fakeSnappyBackend struct {
 
 	linkSnapFailTrigger     string
 	copySnapDataFailTrigger string
+	emptyContainer          snap.Container
 }
 
 func (f *fakeSnappyBackend) OpenSnapFile(snapFilePath string, si *snap.SideInfo) (*snap.Info, snap.Container, error) {
@@ -286,8 +356,13 @@ func (f *fakeSnappyBackend) OpenSnapFile(snapFilePath string, si *snap.SideInfo)
 		op.sinfo = *si
 	}
 
+	name := filepath.Base(snapFilePath)
+	if idx := strings.IndexByte(name, '_'); idx > -1 {
+		name = name[:idx]
+	}
+
 	f.ops = append(f.ops, op)
-	return &snap.Info{Architectures: []string{"all"}}, nil, nil
+	return &snap.Info{SuggestedName: name, Architectures: []string{"all"}}, f.emptyContainer, nil
 }
 
 func (f *fakeSnappyBackend) SetupSnap(snapFilePath string, si *snap.SideInfo, p progress.Meter) error {
@@ -313,34 +388,47 @@ func (f *fakeSnappyBackend) ReadInfo(name string, si *snap.SideInfo) (*snap.Info
 		SuggestedName: name,
 		SideInfo:      *si,
 		Architectures: []string{"all"},
+		Type:          snap.TypeApp,
 	}
-	info.Type = snap.TypeApp
-	if name == "gadget" {
+	if strings.Contains(name, "alias-snap") {
+		name = "alias-snap"
+	}
+	switch name {
+	case "gadget":
 		info.Type = snap.TypeGadget
-	}
-	if name == "core" {
+	case "core":
 		info.Type = snap.TypeOS
-	}
-	if name == "alias-snap" {
+	case "services-snap":
+		var err error
+		info, err = snap.InfoFromSnapYaml([]byte(`name: services-snap
+apps:
+  svc1:
+    daemon: simple
+  svc2:
+    daemon: simple
+`))
+		if err != nil {
+			panic(err)
+		}
+		info.SideInfo = *si
+	case "alias-snap":
 		var err error
 		info, err = snap.InfoFromSnapYaml([]byte(`name: alias-snap
 apps:
   cmd1:
-    aliases: [alias1, alias1.cmd1]
   cmd2:
-    aliases: [alias2]
   cmd3:
-    aliases: [alias3]
   cmd4:
-    aliases: [alias4]
   cmd5:
-    aliases: [alias5]
+  cmddaemon:
+    daemon: simple
 `))
 		if err != nil {
 			panic(err)
 		}
 		info.SideInfo = *si
 	}
+
 	return info, nil
 }
 
@@ -398,18 +486,28 @@ func (f *fakeSnappyBackend) LinkSnap(info *snap.Info) error {
 	return nil
 }
 
-func (f *fakeSnappyBackend) StartSnapServices(info *snap.Info, meter progress.Meter) error {
+func svcSnapMountDir(svcs []*snap.AppInfo) string {
+	if len(svcs) == 0 {
+		return "<no services>"
+	}
+	if svcs[0].Snap == nil {
+		return "<snapless service>"
+	}
+	return svcs[0].Snap.MountDir()
+}
+
+func (f *fakeSnappyBackend) StartServices(svcs []*snap.AppInfo, meter progress.Meter) error {
 	f.ops = append(f.ops, fakeOp{
 		op:   "start-snap-services",
-		name: info.MountDir(),
+		name: svcSnapMountDir(svcs),
 	})
 	return nil
 }
 
-func (f *fakeSnappyBackend) StopSnapServices(info *snap.Info, meter progress.Meter) error {
+func (f *fakeSnappyBackend) StopServices(svcs []*snap.AppInfo, reason snap.ServiceStopReason, meter progress.Meter) error {
 	f.ops = append(f.ops, fakeOp{
-		op:   "stop-snap-services",
-		name: info.MountDir(),
+		op:   fmt.Sprintf("stop-snap-services:%s", reason),
+		name: svcSnapMountDir(svcs),
 	})
 	return nil
 }
@@ -511,23 +609,23 @@ func (f *fakeSnappyBackend) ForeignTask(kind string, status state.Status, snapsu
 	})
 }
 
-func (f *fakeSnappyBackend) MatchingAliases(aliases []*backend.Alias) ([]*backend.Alias, error) {
-	f.ops = append(f.ops, fakeOp{
-		op:      "matching-aliases",
-		aliases: aliases,
-	})
-	return aliases, nil
-}
+type byAlias []*backend.Alias
 
-func (f *fakeSnappyBackend) MissingAliases(aliases []*backend.Alias) ([]*backend.Alias, error) {
-	f.ops = append(f.ops, fakeOp{
-		op:      "missing-aliases",
-		aliases: aliases,
-	})
-	return aliases, nil
+func (ba byAlias) Len() int      { return len(ba) }
+func (ba byAlias) Swap(i, j int) { ba[i], ba[j] = ba[j], ba[i] }
+func (ba byAlias) Less(i, j int) bool {
+	return ba[i].Name < ba[j].Name
 }
 
 func (f *fakeSnappyBackend) UpdateAliases(add []*backend.Alias, remove []*backend.Alias) error {
+	if len(add) != 0 {
+		add = append([]*backend.Alias(nil), add...)
+		sort.Sort(byAlias(add))
+	}
+	if len(remove) != 0 {
+		remove = append([]*backend.Alias(nil), remove...)
+		sort.Sort(byAlias(remove))
+	}
 	f.ops = append(f.ops, fakeOp{
 		op:        "update-aliases",
 		aliases:   add,
