@@ -57,6 +57,7 @@ func (m *SnapManager) AddForeignTaskHandlers(tracker ForeignTaskTracker) {
 		return nil
 	}
 	m.runner.AddHandler("setup-profiles", fakeHandler, fakeHandler)
+	m.runner.AddHandler("auto-connect", fakeHandler, nil)
 	m.runner.AddHandler("remove-profiles", fakeHandler, fakeHandler)
 	m.runner.AddHandler("discard-conns", fakeHandler, fakeHandler)
 	m.runner.AddHandler("validate-snap", fakeHandler, nil)
@@ -71,6 +72,10 @@ func (m *SnapManager) AddForeignTaskHandlers(tracker ForeignTaskTracker) {
 	m.runner.AddHandler("run-hook", func(task *state.Task, _ *tomb.Tomb) error {
 		return nil
 	}, nil)
+	m.runner.AddHandler("configure-snapd", func(t *state.Task, _ *tomb.Tomb) error {
+		return nil
+	}, nil)
+
 }
 
 // AddAdhocTaskHandlers registers handlers for ad hoc test handler
@@ -96,23 +101,21 @@ func MockErrtrackerReport(mock func(string, string, string, map[string]string) (
 	return func() { errtrackerReport = prev }
 }
 
-func MockRefreshInterval(newMinRefreshInterval, newRefreshRandomness time.Duration) (restore func()) {
-	prevMinRefreshInterval := minRefreshInterval
-	prevDefaultRefreshRandomness := defaultRefreshRandomness
-	minRefreshInterval = newMinRefreshInterval
-	defaultRefreshRandomness = newRefreshRandomness
-	return func() {
-		minRefreshInterval = prevMinRefreshInterval
-		defaultRefreshRandomness = prevDefaultRefreshRandomness
-	}
+func MockPrerequisitesRetryTimeout(d time.Duration) (restore func()) {
+	old := prerequisitesRetryTimeout
+	prerequisitesRetryTimeout = d
+	return func() { prerequisitesRetryTimeout = old }
 }
 
 var (
-	CheckSnap            = checkSnap
-	CanRemove            = canRemove
-	CanDisable           = canDisable
-	CachedStore          = cachedStore
-	NameAndRevnoFromSnap = nameAndRevnoFromSnap
+	CheckSnap              = checkSnap
+	CanRemove              = canRemove
+	CanDisable             = canDisable
+	CachedStore            = cachedStore
+	DefaultRefreshSchedule = defaultRefreshSchedule
+	NameAndRevnoFromSnap   = nameAndRevnoFromSnap
+	DoInstall              = doInstall
+	UserFromUserID         = userFromUserID
 )
 
 func PreviousSideInfo(snapst *SnapState) *snap.SideInfo {
@@ -122,7 +125,41 @@ func PreviousSideInfo(snapst *SnapState) *snap.SideInfo {
 // aliases v2
 var (
 	ApplyAliasesChange    = applyAliasesChange
-	AutoAliasesDeltaV2    = autoAliasesDeltaV2
+	AutoAliasesDelta      = autoAliasesDelta
 	RefreshAliases        = refreshAliases
 	CheckAliasesConflicts = checkAliasesConflicts
+	DisableAliases        = disableAliases
 )
+
+// readme files
+var (
+	WriteSnapReadme = writeSnapReadme
+	SnapReadme      = snapReadme
+)
+
+// refreshes
+var (
+	NewAutoRefresh    = newAutoRefresh
+	NewRefreshHints   = newRefreshHints
+	NewCatalogRefresh = newCatalogRefresh
+)
+
+func MockNextRefresh(ar *autoRefresh, when time.Time) {
+	ar.nextRefresh = when
+}
+
+func MockLastRefreshSchedule(ar *autoRefresh, schedule string) {
+	ar.lastRefreshSchedule = schedule
+}
+
+func MockCatalogRefreshNextRefresh(cr *catalogRefresh, when time.Time) {
+	cr.nextCatalogRefresh = when
+}
+
+func MockRefreshRetryDelay(d time.Duration) func() {
+	origRefreshRetryDelay := refreshRetryDelay
+	refreshRetryDelay = d
+	return func() {
+		refreshRetryDelay = origRefreshRetryDelay
+	}
+}
