@@ -21,13 +21,14 @@ package snap
 
 import (
 	"fmt"
+	"os"
 	"sort"
+	"strconv"
 	"strings"
 
 	"gopkg.in/yaml.v2"
 
 	"github.com/snapcore/snapd/strutil"
-	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timeout"
 )
 
@@ -37,17 +38,21 @@ type snapYaml struct {
 	Type             Type                   `yaml:"type"`
 	Architectures    []string               `yaml:"architectures,omitempty"`
 	Assumes          []string               `yaml:"assumes"`
+	Title            string                 `yaml:"title"`
 	Description      string                 `yaml:"description"`
 	Summary          string                 `yaml:"summary"`
+	License          string                 `yaml:"license,omitempty"`
 	LicenseAgreement string                 `yaml:"license-agreement,omitempty"`
 	LicenseVersion   string                 `yaml:"license-version,omitempty"`
-	Epoch            string                 `yaml:"epoch,omitempty"`
+	Epoch            Epoch                  `yaml:"epoch,omitempty"`
+	Base             string                 `yaml:"base,omitempty"`
 	Confinement      ConfinementType        `yaml:"confinement,omitempty"`
 	Environment      strutil.OrderedMap     `yaml:"environment,omitempty"`
 	Plugs            map[string]interface{} `yaml:"plugs,omitempty"`
 	Slots            map[string]interface{} `yaml:"slots,omitempty"`
 	Apps             map[string]appYaml     `yaml:"apps,omitempty"`
 	Hooks            map[string]hookYaml    `yaml:"hooks,omitempty"`
+	Layout           map[string]layoutYaml  `yaml:"layout,omitempty"`
 }
 
 type appYaml struct {
@@ -61,18 +66,43 @@ type appYaml struct {
 	ReloadCommand   string          `yaml:"reload-command,omitempty"`
 	PostStopCommand string          `yaml:"post-stop-command,omitempty"`
 	StopTimeout     timeout.Timeout `yaml:"stop-timeout,omitempty"`
+	Completer       string          `yaml:"completer,omitempty"`
+	RefreshMode     string          `yaml:"refresh-mode,omitempty"`
 
-	RestartCond systemd.RestartCondition `yaml:"restart-condition,omitempty"`
-	SlotNames   []string                 `yaml:"slots,omitempty"`
-	PlugNames   []string                 `yaml:"plugs,omitempty"`
+	RestartCond RestartCondition `yaml:"restart-condition,omitempty"`
+	SlotNames   []string         `yaml:"slots,omitempty"`
+	PlugNames   []string         `yaml:"plugs,omitempty"`
 
 	BusName string `yaml:"bus-name,omitempty"`
 
 	Environment strutil.OrderedMap `yaml:"environment,omitempty"`
+
+	Sockets map[string]socketsYaml `yaml:"sockets,omitempty"`
+
+	After  []string `yaml:"after,omitempty"`
+	Before []string `yaml:"before,omitempty"`
+
+	Timer string `yaml:"timer,omitempty"`
 }
 
 type hookYaml struct {
 	PlugNames []string `yaml:"plugs,omitempty"`
+	SlotNames []string `yaml:"slots,omitempty"`
+}
+
+type layoutYaml struct {
+	Bind     string `yaml:"bind,omitempty"`
+	BindFile string `yaml:"bind-file,omitempty"`
+	Type     string `yaml:"type,omitempty"`
+	User     string `yaml:"user,omitempty"`
+	Group    string `yaml:"group,omitempty"`
+	Mode     string `yaml:"mode,omitempty"`
+	Symlink  string `yaml:"symlink,omitempty"`
+}
+
+type socketsYaml struct {
+	ListenStream string      `yaml:"listen-stream,omitempty"`
+	SocketMode   os.FileMode `yaml:"socket-mode,omitempty"`
 }
 
 // InfoFromSnapYaml creates a new info based on the given snap.yaml data
@@ -115,12 +145,43 @@ func InfoFromSnapYaml(yamlData []byte) (*Info, error) {
 	// Bind unbound plugs to all apps and hooks
 	bindUnboundPlugs(globalPlugNames, snap)
 
-	// Bind unbound slots to all apps
+	// Bind unbound slots to all apps and hooks
 	bindUnboundSlots(globalSlotNames, snap)
 
+	// Collect layout elements.
+	if y.Layout != nil {
+		snap.Layout = make(map[string]*Layout, len(y.Layout))
+		for path, l := range y.Layout {
+			var mode os.FileMode = 0755
+			if l.Mode != "" {
+				m, err := strconv.ParseUint(l.Mode, 8, 32)
+				if err != nil {
+					return nil, err
+				}
+				mode = os.FileMode(m)
+			}
+			user := "root"
+			if l.User != "" {
+				user = l.User
+			}
+			group := "root"
+			if l.Group != "" {
+				group = l.Group
+			}
+			snap.Layout[path] = &Layout{
+				Snap: snap, Path: path,
+				Bind: l.Bind, Type: l.Type, Symlink: l.Symlink, BindFile: l.BindFile,
+				User: user, Group: group, Mode: mode,
+			}
+		}
+	}
+
 	// Rename specific plugs on the core snap.
 	snap.renameClashingCorePlugs()
 
+	snap.BadInterfaces = make(map[string]string)
+	SanitizePlugsSlots(snap)
+
 	// FIXME: validation of the fields
 	return snap, nil
 }
@@ -137,10 +198,6 @@ func infoSkeletonFromSnapYaml(y snapYaml) *Info {
 	if y.Type != "" {
 		typ = y.Type
 	}
-	epoch := "0"
-	if y.Epoch != "" {
-		epoch = y.Epoch
-	}
 	confinement := StrictConfinement
 	if y.Confinement != "" {
 		confinement = y.Confinement
@@ -153,14 +210,17 @@ func infoSkeletonFromSnapYaml(y snapYaml) *Info {
 		Type:                typ,
 		Architectures:       architectures,
 		Assumes:             y.Assumes,
+		OriginalTitle:       y.Title,
 		OriginalDescription: y.Description,
 		OriginalSummary:     y.Summary,
+		License:             y.License,
 		LicenseAgreement:    y.LicenseAgreement,
 		LicenseVersion:      y.LicenseVersion,
-		Epoch:               epoch,
+		Epoch:               y.Epoch,
 		Confinement:         confinement,
+		Base:                y.Base,
 		Apps:                make(map[string]*AppInfo),
-		Aliases:             make(map[string]*AppInfo),
+		LegacyAliases:       make(map[string]*AppInfo),
 		Hooks:               make(map[string]*HookInfo),
 		Plugs:               make(map[string]*PlugInfo),
 		Slots:               make(map[string]*SlotInfo),
@@ -212,6 +272,9 @@ func setSlotsFromSnapYaml(y snapYaml, snap *Info) error {
 		if len(y.Apps) > 0 {
 			snap.Slots[name].Apps = make(map[string]*AppInfo)
 		}
+		if len(y.Hooks) > 0 {
+			snap.Slots[name].Hooks = make(map[string]*HookInfo)
+		}
 	}
 
 	return nil
@@ -223,7 +286,7 @@ func setAppsFromSnapYaml(y snapYaml, snap *Info) error {
 		app := &AppInfo{
 			Snap:            snap,
 			Name:            appName,
-			Aliases:         yApp.Aliases,
+			LegacyAliases:   yApp.Aliases,
 			Command:         yApp.Command,
 			Daemon:          yApp.Daemon,
 			StopTimeout:     yApp.StopTimeout,
@@ -233,6 +296,10 @@ func setAppsFromSnapYaml(y snapYaml, snap *Info) error {
 			RestartCond:     yApp.RestartCond,
 			BusName:         yApp.BusName,
 			Environment:     yApp.Environment,
+			Completer:       yApp.Completer,
+			RefreshMode:     yApp.RefreshMode,
+			Before:          yApp.Before,
+			After:           yApp.After,
 		}
 		if len(y.Plugs) > 0 || len(yApp.PlugNames) > 0 {
 			app.Plugs = make(map[string]*PlugInfo)
@@ -240,12 +307,16 @@ func setAppsFromSnapYaml(y snapYaml, snap *Info) error {
 		if len(y.Slots) > 0 || len(yApp.SlotNames) > 0 {
 			app.Slots = make(map[string]*SlotInfo)
 		}
+		if len(yApp.Sockets) > 0 {
+			app.Sockets = make(map[string]*SocketInfo, len(yApp.Sockets))
+		}
+
 		snap.Apps[appName] = app
-		for _, alias := range app.Aliases {
-			if snap.Aliases[alias] != nil {
-				return fmt.Errorf("cannot set %q as alias for both %q and %q", alias, snap.Aliases[alias].Name, appName)
+		for _, alias := range app.LegacyAliases {
+			if snap.LegacyAliases[alias] != nil {
+				return fmt.Errorf("cannot set %q as alias for both %q and %q", alias, snap.LegacyAliases[alias].Name, appName)
 			}
-			snap.Aliases[alias] = app
+			snap.LegacyAliases[alias] = app
 		}
 		// Bind all plugs/slots listed in this app
 		for _, plugName := range yApp.PlugNames {
@@ -277,6 +348,20 @@ func setAppsFromSnapYaml(y snapYaml, snap *Info) error {
 			app.Slots[slotName] = slot
 			slot.Apps[appName] = app
 		}
+		for name, data := range yApp.Sockets {
+			app.Sockets[name] = &SocketInfo{
+				App:          app,
+				Name:         name,
+				ListenStream: data.ListenStream,
+				SocketMode:   data.SocketMode,
+			}
+		}
+		if yApp.Timer != "" {
+			app.Timer = &TimerInfo{
+				App:   app,
+				Timer: yApp.Timer,
+			}
+		}
 	}
 	return nil
 }
@@ -295,6 +380,10 @@ func setHooksFromSnapYaml(y snapYaml, snap *Info) {
 		if len(y.Plugs) > 0 || len(yHook.PlugNames) > 0 {
 			hook.Plugs = make(map[string]*PlugInfo)
 		}
+		if len(y.Slots) > 0 || len(yHook.SlotNames) > 0 {
+			hook.Slots = make(map[string]*SlotInfo)
+		}
+
 		snap.Hooks[hookName] = hook
 		// Bind all plugs/slots listed in this hook
 		for _, plugName := range yHook.PlugNames {
@@ -314,6 +403,23 @@ func setHooksFromSnapYaml(y snapYaml, snap *Info) {
 			hook.Plugs[plugName] = plug
 			plug.Hooks[hookName] = hook
 		}
+		for _, slotName := range yHook.SlotNames {
+			slot, ok := snap.Slots[slotName]
+			if !ok {
+				// Create implicit slot definitions if required
+				slot = &SlotInfo{
+					Snap:      snap,
+					Name:      slotName,
+					Interface: slotName,
+					Hooks:     make(map[string]*HookInfo),
+				}
+				snap.Slots[slotName] = slot
+			} else if slot.Hooks == nil {
+				slot.Hooks = make(map[string]*HookInfo)
+			}
+			hook.Slots[slotName] = slot
+			slot.Hooks[hookName] = hook
+		}
 	}
 }
 
@@ -349,11 +455,17 @@ func bindUnboundSlots(slotNames []string, snap *Info) error {
 			return fmt.Errorf("no slot named %q", slotName)
 		}
 
-		if len(slot.Apps) == 0 {
+		// A slot is considered unbound if it isn't being used by any apps
+		// or hooks. In which case we bind them to all apps and hooks.
+		if len(slot.Apps) == 0 && len(slot.Hooks) == 0 {
 			for appName, app := range snap.Apps {
 				app.Slots[slotName] = slot
 				slot.Apps[appName] = app
 			}
+			for hookName, hook := range snap.Hooks {
+				hook.Slots[slotName] = slot
+				slot.Hooks[hookName] = hook
+			}
 		}
 	}
 
@@ -400,7 +512,7 @@ func convertToSlotOrPlugData(plugOrSlot, name string, data interface{}) (iface,
 				if attrs == nil {
 					attrs = make(map[string]interface{})
 				}
-				value, err := validateAttr(valueData)
+				value, err := normalizeYamlValue(valueData)
 				if err != nil {
 					return "", "", nil, fmt.Errorf("attribute %q of %s %q: %v", key, plugOrSlot, name, err)
 				}
@@ -414,8 +526,8 @@ func convertToSlotOrPlugData(plugOrSlot, name string, data interface{}) (iface,
 	}
 }
 
-// validateAttr validates an attribute value and returns a normalized version of it (map[interface{}]interface{} is turned into map[string]interface{})
-func validateAttr(v interface{}) (interface{}, error) {
+// normalizeYamlValue validates values and returns a normalized version of it (map[interface{}]interface{} is turned into map[string]interface{})
+func normalizeYamlValue(v interface{}) (interface{}, error) {
 	switch x := v.(type) {
 	case string:
 		return x, nil
@@ -425,10 +537,14 @@ func validateAttr(v interface{}) (interface{}, error) {
 		return int64(x), nil
 	case int64:
 		return x, nil
+	case float64:
+		return x, nil
+	case float32:
+		return float64(x), nil
 	case []interface{}:
 		l := make([]interface{}, len(x))
 		for i, el := range x {
-			el, err := validateAttr(el)
+			el, err := normalizeYamlValue(el)
 			if err != nil {
 				return nil, err
 			}
@@ -440,16 +556,26 @@ func validateAttr(v interface{}) (interface{}, error) {
 		for k, item := range x {
 			kStr, ok := k.(string)
 			if !ok {
-				return nil, fmt.Errorf("non-string key in attribute map: %v", k)
+				return nil, fmt.Errorf("non-string key: %v", k)
 			}
-			item, err := validateAttr(item)
+			item, err := normalizeYamlValue(item)
 			if err != nil {
 				return nil, err
 			}
 			m[kStr] = item
 		}
 		return m, nil
+	case map[string]interface{}:
+		m := make(map[string]interface{}, len(x))
+		for k, item := range x {
+			item, err := normalizeYamlValue(item)
+			if err != nil {
+				return nil, err
+			}
+			m[k] = item
+		}
+		return m, nil
 	default:
-		return nil, fmt.Errorf("invalid attribute scalar: %v", v)
+		return nil, fmt.Errorf("invalid scalar: %v", v)
 	}
 }
