@@ -28,14 +28,17 @@ import (
 
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/strutil"
-	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timeout"
+
+	"github.com/snapcore/snapd/testutil"
 )
 
 // Hook up check.v1 into the "go test" runner
 func Test(t *testing.T) { TestingT(t) }
 
-type InfoSnapYamlTestSuite struct{}
+type InfoSnapYamlTestSuite struct {
+	testutil.BaseTest
+}
 
 var _ = Suite(&InfoSnapYamlTestSuite{})
 
@@ -44,6 +47,15 @@ version: 1.0
 type: app
 `)
 
+func (s *InfoSnapYamlTestSuite) SetUpTest(c *C) {
+	s.BaseTest.SetUpTest(c)
+	s.BaseTest.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
+}
+
+func (s *InfoSnapYamlTestSuite) TearDownTest(c *C) {
+	s.BaseTest.TearDownTest(c)
+}
+
 func (s *InfoSnapYamlTestSuite) TestSimple(c *C) {
 	info, err := snap.InfoFromSnapYaml(mockYaml)
 	c.Assert(err, IsNil)
@@ -59,16 +71,20 @@ func (s *InfoSnapYamlTestSuite) TestFail(c *C) {
 
 type YamlSuite struct {
 	restore func()
+	testutil.BaseTest
 }
 
 var _ = Suite(&YamlSuite{})
 
 func (s *YamlSuite) SetUpTest(c *C) {
+	s.BaseTest.SetUpTest(c)
+	s.BaseTest.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
 	hookType := snap.NewHookType(regexp.MustCompile(".*"))
 	s.restore = snap.MockSupportedHookTypes([]*snap.HookType{hookType})
 }
 
 func (s *YamlSuite) TearDownTest(c *C) {
+	s.BaseTest.TearDownTest(c)
 	s.restore()
 }
 
@@ -449,7 +465,7 @@ plugs:
         interface: serial-port
         foo: null
 `))
-	c.Assert(err, ErrorMatches, `attribute "foo" of plug \"serial\": invalid attribute scalar:.*`)
+	c.Assert(err, ErrorMatches, `attribute "foo" of plug \"serial\": invalid scalar:.*`)
 }
 
 func (s *YamlSuite) TestUnmarshalInvalidAttributeMapKey(c *C) {
@@ -463,7 +479,7 @@ plugs:
           baz:
           - 1: A
 `))
-	c.Assert(err, ErrorMatches, `attribute "bar" of plug \"serial\": non-string key in attribute map: 1`)
+	c.Assert(err, ErrorMatches, `attribute "bar" of plug \"serial\": non-string key: 1`)
 }
 
 // Tests focusing on slots
@@ -726,6 +742,73 @@ slots:
 	})
 }
 
+func (s *YamlSuite) TestUnmarshalGlobalSlotsBindToHooks(c *C) {
+	// NOTE: yaml content cannot use tabs, indent the section with spaces.
+	info, err := snap.InfoFromSnapYaml([]byte(`
+name: snap
+slots:
+    test-slot:
+hooks:
+    test-hook:
+`))
+	c.Assert(err, IsNil)
+	c.Check(info.Name(), Equals, "snap")
+	c.Check(info.Plugs, HasLen, 0)
+	c.Check(info.Slots, HasLen, 1)
+	c.Check(info.Apps, HasLen, 0)
+	c.Check(info.Hooks, HasLen, 1)
+
+	slot, ok := info.Slots["test-slot"]
+	c.Assert(ok, Equals, true, Commentf("Expected slots to include 'test-slot'"))
+	hook, ok := info.Hooks["test-hook"]
+	c.Assert(ok, Equals, true, Commentf("Expected hooks to include 'test-hook'"))
+
+	c.Check(slot, DeepEquals, &snap.SlotInfo{
+		Snap:      info,
+		Name:      "test-slot",
+		Interface: "test-slot",
+		Hooks:     map[string]*snap.HookInfo{hook.Name: hook},
+	})
+	c.Check(hook, DeepEquals, &snap.HookInfo{
+		Snap:  info,
+		Name:  "test-hook",
+		Slots: map[string]*snap.SlotInfo{slot.Name: slot},
+	})
+}
+
+func (s *YamlSuite) TestUnmarshalHookWithSlot(c *C) {
+	// NOTE: yaml content cannot use tabs, indent the section with spaces.
+	info, err := snap.InfoFromSnapYaml([]byte(`
+name: snap
+hooks:
+    test-hook:
+        slots: [test-slot]
+`))
+	c.Assert(err, IsNil)
+	c.Check(info.Name(), Equals, "snap")
+	c.Check(info.Plugs, HasLen, 0)
+	c.Check(info.Slots, HasLen, 1)
+	c.Check(info.Apps, HasLen, 0)
+	c.Check(info.Hooks, HasLen, 1)
+
+	slot, ok := info.Slots["test-slot"]
+	c.Assert(ok, Equals, true, Commentf("Expected slots to include 'test-slot'"))
+	hook, ok := info.Hooks["test-hook"]
+	c.Assert(ok, Equals, true, Commentf("Expected hooks to include 'test-hook'"))
+
+	c.Check(slot, DeepEquals, &snap.SlotInfo{
+		Snap:      info,
+		Name:      "test-slot",
+		Interface: "test-slot",
+		Hooks:     map[string]*snap.HookInfo{hook.Name: hook},
+	})
+	c.Check(hook, DeepEquals, &snap.HookInfo{
+		Snap:  info,
+		Name:  "test-hook",
+		Slots: map[string]*snap.SlotInfo{slot.Name: slot},
+	})
+}
+
 func (s *YamlSuite) TestUnmarshalCorruptedSlotWithNonStringInterfaceName(c *C) {
 	// NOTE: yaml content cannot use tabs, indent the section with spaces.
 	_, err := snap.InfoFromSnapYaml([]byte(`
@@ -791,7 +874,7 @@ slots:
         interface: serial-port
         foo: null
 `))
-	c.Assert(err, ErrorMatches, `attribute "foo" of slot \"serial\": invalid attribute scalar:.*`)
+	c.Assert(err, ErrorMatches, `attribute "foo" of slot \"serial\": invalid scalar:.*`)
 }
 
 func (s *YamlSuite) TestUnmarshalHook(c *C) {
@@ -1053,10 +1136,12 @@ func (s *YamlSuite) TestUnmarshalComplexExample(c *C) {
 	info, err := snap.InfoFromSnapYaml([]byte(`
 name: foo
 version: 1.2
+title: Foo
 summary: foo app
 type: app
 epoch: 1*
 confinement: devmode
+license: GPL-3.0
 description: |
     Foo provides useful services
 apps:
@@ -1070,6 +1155,7 @@ apps:
 hooks:
     test-hook:
        plugs: [foo-socket-plug]
+       slots: [foo-socket-slot]
 plugs:
     foo-socket-plug:
         interface: socket
@@ -1088,8 +1174,9 @@ slots:
 	c.Check(info.Name(), Equals, "foo")
 	c.Check(info.Version, Equals, "1.2")
 	c.Check(info.Type, Equals, snap.TypeApp)
-	c.Check(info.Epoch, Equals, "1*")
+	c.Check(info.Epoch.String(), Equals, "1*")
 	c.Check(info.Confinement, Equals, snap.DevModeConfinement)
+	c.Check(info.Title(), Equals, "Foo")
 	c.Check(info.Summary(), Equals, "foo app")
 	c.Check(info.Description(), Equals, "Foo provides useful services\n")
 	c.Check(info.Apps, HasLen, 2)
@@ -1099,6 +1186,7 @@ slots:
 	c.Check(info.Publisher, Equals, "")
 	c.Check(info.PublisherID, Equals, "")
 	c.Check(info.Channel, Equals, "")
+	c.Check(info.License, Equals, "GPL-3.0")
 
 	app1 := info.Apps["daemon"]
 	app2 := info.Apps["foo"]
@@ -1136,14 +1224,16 @@ slots:
 	c.Check(app2.Slots, DeepEquals, map[string]*snap.SlotInfo{
 		slot2.Name: slot2})
 
-	// hook1 has two plugs ("foo-socket", "logging"). The plug "logging" is
-	// global while "foo-socket" is hook-bound.
+	// hook1 has two plugs ("foo-socket", "logging") and two slots ("foo-socket", "tracing").
+	// The plug "logging" and slot "tracing" are global while "foo-socket" is hook-bound.
 
 	c.Assert(hook, NotNil)
 	c.Check(hook.Snap, Equals, info)
 	c.Check(hook.Name, Equals, "test-hook")
 	c.Check(hook.Plugs, DeepEquals, map[string]*snap.PlugInfo{
 		plug3.Name: plug3, plug4.Name: plug4})
+	c.Check(hook.Slots, DeepEquals, map[string]*snap.SlotInfo{
+		slot1.Name: slot1, slot2.Name: slot2})
 
 	// plug1 ("network") is implicitly defined and app-bound to "daemon"
 
@@ -1226,7 +1316,7 @@ version: 1.0
 `)
 	info, err := snap.InfoFromSnapYaml(y)
 	c.Assert(err, IsNil)
-	c.Assert(info.Epoch, Equals, "0")
+	c.Assert(info.Epoch.String(), Equals, "0")
 }
 
 func (s *YamlSuite) TestSnapYamlConfinementDefault(c *C) {
@@ -1333,24 +1423,84 @@ apps:
    post-stop-command: post-stop-cmd
    restart-condition: on-abnormal
    bus-name: busName
+   sockets:
+     sock1:
+       listen-stream: $SNAP_DATA/sock1.socket
+       socket-mode: 0666
 `)
 	info, err := snap.InfoFromSnapYaml(y)
 	c.Assert(err, IsNil)
+
+	app := snap.AppInfo{
+		Snap:            info,
+		Name:            "svc",
+		Command:         "svc1",
+		Daemon:          "forking",
+		RestartCond:     snap.RestartOnAbnormal,
+		StopTimeout:     timeout.Timeout(25 * time.Second),
+		StopCommand:     "stop-cmd",
+		PostStopCommand: "post-stop-cmd",
+		BusName:         "busName",
+		Sockets:         map[string]*snap.SocketInfo{},
+	}
+
+	app.Sockets["sock1"] = &snap.SocketInfo{
+		App:          &app,
+		Name:         "sock1",
+		ListenStream: "$SNAP_DATA/sock1.socket",
+		SocketMode:   0666,
+	}
+
+	c.Check(info.Apps, DeepEquals, map[string]*snap.AppInfo{"svc": &app})
+}
+
+func (s *YamlSuite) TestDaemonListenStreamAsInteger(c *C) {
+	y := []byte(`name: wat
+version: 42
+apps:
+ svc:
+   command: svc
+   sockets:
+     sock:
+       listen-stream: 8080
+`)
+	info, err := snap.InfoFromSnapYaml(y)
+	c.Assert(err, IsNil)
+
+	app := snap.AppInfo{
+		Snap:    info,
+		Name:    "svc",
+		Command: "svc",
+		Sockets: map[string]*snap.SocketInfo{},
+	}
+
+	app.Sockets["sock"] = &snap.SocketInfo{
+		App:          &app,
+		Name:         "sock",
+		ListenStream: "8080",
+	}
+
 	c.Check(info.Apps, DeepEquals, map[string]*snap.AppInfo{
-		"svc": {
-			Snap:            info,
-			Name:            "svc",
-			Command:         "svc1",
-			Daemon:          "forking",
-			RestartCond:     systemd.RestartOnAbnormal,
-			StopTimeout:     timeout.Timeout(25 * time.Second),
-			StopCommand:     "stop-cmd",
-			PostStopCommand: "post-stop-cmd",
-			BusName:         "busName",
-		},
+		"svc": &app,
 	})
 }
 
+func (s *YamlSuite) TestDaemonInvalidSocketMode(c *C) {
+	y := []byte(`name: wat
+version: 42
+apps:
+ svc:
+   command: svc
+   sockets:
+     sock:
+       listen-stream: 8080
+       socket-mode: asdfasdf
+`)
+	_, err := snap.InfoFromSnapYaml(y)
+	c.Check(err.Error(), Equals, "info failed to parse: yaml: unmarshal errors:\n"+
+		"  line 9: cannot unmarshal !!str `asdfasdf` into os.FileMode")
+}
+
 func (s *YamlSuite) TestSnapYamlGlobalEnvironment(c *C) {
 	y := []byte(`
 name: foo
@@ -1403,10 +1553,10 @@ apps:
 	info, err := snap.InfoFromSnapYaml(y)
 	c.Assert(err, IsNil)
 
-	c.Check(info.Apps["foo"].Aliases, DeepEquals, []string{"foo"})
-	c.Check(info.Apps["bar"].Aliases, DeepEquals, []string{"bar", "bar1"})
+	c.Check(info.Apps["foo"].LegacyAliases, DeepEquals, []string{"foo"})
+	c.Check(info.Apps["bar"].LegacyAliases, DeepEquals, []string{"bar", "bar1"})
 
-	c.Check(info.Aliases, DeepEquals, map[string]*snap.AppInfo{
+	c.Check(info.LegacyAliases, DeepEquals, map[string]*snap.AppInfo{
 		"foo":  info.Apps["foo"],
 		"bar":  info.Apps["bar"],
 		"bar1": info.Apps["bar"],
@@ -1426,3 +1576,97 @@ apps:
 	_, err := snap.InfoFromSnapYaml(y)
 	c.Assert(err, ErrorMatches, `cannot set "bar" as alias for both ("foo" and "bar"|"bar" and "foo")`)
 }
+
+func (s *YamlSuite) TestSnapYamlAppStartOrder(c *C) {
+	y := []byte(`name: wat
+version: 42
+apps:
+ foo:
+   after: [bar, zed]
+ bar:
+   before: [foo]
+ baz:
+   after: [foo]
+ zed:
+
+`)
+	info, err := snap.InfoFromSnapYaml(y)
+	c.Assert(err, IsNil)
+
+	c.Check(info.Apps, DeepEquals, map[string]*snap.AppInfo{
+		"foo": {
+			Snap:  info,
+			Name:  "foo",
+			After: []string{"bar", "zed"},
+		},
+		"bar": {
+			Snap:   info,
+			Name:   "bar",
+			Before: []string{"foo"},
+		},
+		"baz": {
+			Snap:  info,
+			Name:  "baz",
+			After: []string{"foo"},
+		},
+		"zed": {
+			Snap: info,
+			Name: "zed",
+		},
+	})
+}
+
+func (s *YamlSuite) TestLayout(c *C) {
+	y := []byte(`
+name: foo
+version: 1.0
+layout:
+  /usr/share/foo:
+    bind: $SNAP/usr/share/foo
+  /usr/share/bar:
+    symlink: $SNAP/usr/share/bar
+  /etc/froz:
+    bind-file: $SNAP/etc/froz
+`)
+	info, err := snap.InfoFromSnapYaml(y)
+	c.Assert(err, IsNil)
+	c.Assert(info.Layout["/usr/share/foo"], DeepEquals, &snap.Layout{
+		Snap:  info,
+		Path:  "/usr/share/foo",
+		Bind:  "$SNAP/usr/share/foo",
+		User:  "root",
+		Group: "root",
+		Mode:  0755,
+	})
+	c.Assert(info.Layout["/usr/share/bar"], DeepEquals, &snap.Layout{
+		Snap:    info,
+		Path:    "/usr/share/bar",
+		Symlink: "$SNAP/usr/share/bar",
+		User:    "root",
+		Group:   "root",
+		Mode:    0755,
+	})
+	c.Assert(info.Layout["/etc/froz"], DeepEquals, &snap.Layout{
+		Snap:     info,
+		Path:     "/etc/froz",
+		BindFile: "$SNAP/etc/froz",
+		User:     "root",
+		Group:    "root",
+		Mode:     0755,
+	})
+}
+
+func (s *YamlSuite) TestSnapYamlAppTimer(c *C) {
+	y := []byte(`name: wat
+version: 42
+apps:
+ foo:
+   daemon: oneshot
+   timer: mon,10:00-12:00
+
+`)
+	info, err := snap.InfoFromSnapYaml(y)
+	c.Assert(err, IsNil)
+	app := info.Apps["foo"]
+	c.Check(app.Timer, DeepEquals, &snap.TimerInfo{App: app, Timer: "mon,10:00-12:00"})
+}
